/*
 * SonarQube
 * Copyright (C) 2009-2016 SonarSource SA
 * mailto:contact AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package org.sonar.server.platform.platformlevel;

import java.util.Properties;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.sonar.api.utils.System2;
import org.sonar.core.platform.PluginRepository;
import org.sonar.db.charset.DatabaseCharsetChecker;
import org.sonar.process.ProcessProperties;
import org.sonar.server.platform.Platform;
import org.sonar.server.platform.cluster.Cluster;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.mock;

public class PlatformLevel2Test {

  @Rule
  public TemporaryFolder tempFolder = new TemporaryFolder();

  private Properties props = new Properties();

  @Before
  public void setUp() throws Exception {
    // these are mandatory settings declared by bootstrap process
    props.setProperty(ProcessProperties.PATH_HOME, tempFolder.newFolder().getAbsolutePath());
    props.setProperty(ProcessProperties.PATH_DATA, tempFolder.newFolder().getAbsolutePath());
    props.setProperty(ProcessProperties.PATH_TEMP, tempFolder.newFolder().getAbsolutePath());
  }

  @Test
  public void add_all_components_by_default() {
    PlatformLevel1 level1 = new PlatformLevel1(mock(Platform.class), props);
    level1.configureLevel();

    PlatformLevel2 underTest = new PlatformLevel2(level1);
    underTest.configureLevel();

    // some level1 components
    assertThat(underTest.getOptional(Cluster.class)).isPresent();
    assertThat(underTest.getOptional(System2.class)).isPresent();

    // level2 component that does not depend on cluster state
    assertThat(underTest.getOptional(PluginRepository.class)).isPresent();

    // level2 component that is injected only on "startup leaders"
    assertThat(underTest.getOptional(DatabaseCharsetChecker.class)).isPresent();
  }

  @Test
  public void do_not_add_all_components_when_startup_follower() {
    props.setProperty("sonar.cluster.enabled", "true");
    PlatformLevel1 level1 = new PlatformLevel1(mock(Platform.class), props);
    level1.configureLevel();

    PlatformLevel2 underTest = new PlatformLevel2(level1);
    underTest.configureLevel();

    assertThat(underTest.get(Cluster.class).isStartupLeader()).isFalse();

    // level2 component that does not depend on cluster state
    assertThat(underTest.getOptional(PluginRepository.class)).isPresent();

    // level2 component that is injected only on "startup leaders"
    assertThat(underTest.getOptional(DatabaseCharsetChecker.class)).isNotPresent();
  }
}
